@@ -27,12 +27,11 @@
 #![deny(missing_docs)]
 #![warn(noop_method_call)]
 #![deny(unreachable_pub)]
-#![deny(clippy::all)]
+#![warn(clippy::all)]
 #![deny(clippy::await_holding_lock)]
 #![deny(clippy::cargo_common_metadata)]
 #![deny(clippy::cast_lossless)]
 #![deny(clippy::checked_conversions)]
-#![warn(clippy::clone_on_ref_ptr)]
 #![warn(clippy::cognitive_complexity)]
 #![deny(clippy::debug_assert_with_mut_call)]
 #![deny(clippy::exhaustive_enums)]
@@ -57,7 +56,7 @@
 
 pub mod authority;
 mod bootstrap;
-mod config;
+pub mod config;
 mod docid;
 mod docmeta;
 mod err;
@@ -67,36 +66,100 @@ mod shared_ref;
 mod state;
 mod storage;
 
+#[cfg(feature = "dirfilter")]
+pub mod filter;
+
 use crate::docid::{CacheUsage, ClientRequest, DocQuery};
+#[cfg(not(feature = "experimental-api"))]
 use crate::shared_ref::SharedMutArc;
-use crate::storage::sqlite::SqliteStore;
+#[cfg(feature = "experimental-api")]
+pub use crate::shared_ref::SharedMutArc;
+use crate::storage::DynStore;
+use postage::watch;
 pub use retry::DownloadSchedule;
 use tor_circmgr::CircMgr;
-use tor_netdir::NetDir;
+use tor_netdir::{DirEvent, NetDir, NetDirProvider};
 use tor_netdoc::doc::netstatus::ConsensusFlavor;
 
-use futures::{channel::oneshot, task::SpawnExt};
+use async_trait::async_trait;
+use futures::{channel::oneshot, stream::BoxStream, task::SpawnExt};
 use tor_rtcompat::{Runtime, SleepProviderExt};
-use tracing::{info, trace, warn};
+use tracing::{debug, info, trace, warn};
 
+use std::sync::atomic::{AtomicBool, Ordering};
 use std::sync::{Arc, Mutex};
+use std::time::Duration;
 use std::{collections::HashMap, sync::Weak};
 use std::{fmt::Debug, time::SystemTime};
 
 pub use authority::{Authority, AuthorityBuilder};
 pub use config::{
-    DirMgrConfig, DirMgrConfigBuilder, DownloadScheduleConfig, DownloadScheduleConfigBuilder,
-    NetworkConfig, NetworkConfigBuilder,
+    DirMgrConfig, DownloadScheduleConfig, DownloadScheduleConfigBuilder, NetworkConfig,
+    NetworkConfigBuilder,
 };
 pub use docid::DocId;
 pub use err::Error;
-pub use event::DirEvent;
+pub use event::{DirBootstrapEvents, DirBootstrapStatus, DirStatus};
 pub use storage::DocumentText;
-pub use tor_netdir::fallback::{FallbackDir, FallbackDirBuilder};
+pub use tor_guardmgr::fallback::{FallbackDir, FallbackDirBuilder};
 
 /// A Result as returned by this crate.
 pub type Result<T> = std::result::Result<T, Error>;
 
+/// Trait for DirMgr implementations
+#[async_trait]
+pub trait DirProvider: NetDirProvider {
+    /// Try to change our configuration to `new_config`.
+    ///
+    /// Actual behavior will depend on the value of `how`.
+    fn reconfigure(
+        &self,
+        new_config: &DirMgrConfig,
+        how: tor_config::Reconfigure,
+    ) -> std::result::Result<(), tor_config::ReconfigureError>;
+
+    /// Bootstrap a `DirProvider` that hasn't been bootstrapped yet.
+    async fn bootstrap(&self) -> Result<()>;
+
+    /// Return a stream of [`DirBootstrapStatus`] events to tell us about changes
+    /// in the latest directory's bootstrap status.
+    ///
+    /// Note that this stream can be lossy: the caller will not necessarily
+    /// observe every event on the stream
+    fn bootstrap_events(&self) -> BoxStream<'static, DirBootstrapStatus>;
+}
+
+// NOTE(eta): We can't implement this for Arc<DirMgr<R>> due to trait coherence rules, so instead
+//            there's a blanket impl for Arc<T> in tor-netdir.
+impl<R: Runtime> NetDirProvider for DirMgr<R> {
+    fn latest_netdir(&self) -> Option<Arc<NetDir>> {
+        self.opt_netdir()
+    }
+
+    fn events(&self) -> BoxStream<'static, DirEvent> {
+        Box::pin(self.events.subscribe())
+    }
+}
+
+#[async_trait]
+impl<R: Runtime> DirProvider for Arc<DirMgr<R>> {
+    fn reconfigure(
+        &self,
+        new_config: &DirMgrConfig,
+        how: tor_config::Reconfigure,
+    ) -> std::result::Result<(), tor_config::ReconfigureError> {
+        DirMgr::reconfigure(self, new_config, how)
+    }
+
+    async fn bootstrap(&self) -> Result<()> {
+        DirMgr::bootstrap(self).await
+    }
+
+    fn bootstrap_events(&self) -> BoxStream<'static, DirBootstrapStatus> {
+        Box::pin(DirMgr::bootstrap_events(self))
+    }
+}
+
 /// A directory manager to download, fetch, and cache a Tor directory.
 ///
 /// A DirMgr can operate in three modes:
@@ -117,7 +180,8 @@ pub struct DirMgr<R: Runtime> {
     /// Handle to our sqlite cache.
     // TODO(nickm): I'd like to use an rwlock, but that's not feasible, since
     // rusqlite::Connection isn't Sync.
-    store: Mutex<SqliteStore>,
+    // TODO is needed?
+    store: Mutex<DynStore>,
     /// Our latest sufficiently bootstrapped directory, if we have one.
     ///
     /// We use the RwLock so that we can give this out to a bunch of other
@@ -127,11 +191,78 @@ pub struct DirMgr<R: Runtime> {
     /// A publisher handle that we notify whenever the consensus changes.
     events: event::FlagPublisher<DirEvent>,
 
+    /// A publisher handle that we notify whenever our bootstrapping status
+    /// changes.
+    send_status: Mutex<watch::Sender<event::DirBootstrapStatus>>,
+
+    /// A receiver handle that gets notified whenever our bootstrapping status
+    /// changes.
+    ///
+    /// We don't need to keep this drained, since `postage::watch` already knows
+    /// to discard unread events.
+    receive_status: DirBootstrapEvents,
+
     /// A circuit manager, if this DirMgr supports downloading.
     circmgr: Option<Arc<CircMgr<R>>>,
 
     /// Our asynchronous runtime.
     runtime: R,
+
+    /// Whether or not we're operating in offline mode.
+    offline: bool,
+
+    /// If we're not in offline mode, stores whether or not the `DirMgr` has attempted
+    /// to bootstrap yet or not.
+    ///
+    /// This exists in order to prevent starting two concurrent bootstrap tasks.
+    ///
+    /// (In offline mode, this does nothing.)
+    bootstrap_started: AtomicBool,
+
+    /// A filter that gets applied to directory objects before we use them.
+    #[cfg(feature = "dirfilter")]
+    filter: crate::filter::FilterConfig,
+}
+
+/// RAII guard to reset an AtomicBool on drop.
+struct BoolResetter<'a> {
+    /// The bool to reset.
+    inner: &'a AtomicBool,
+    /// What value to store.
+    reset_to: bool,
+    /// What atomic ordering to use.
+    ordering: Ordering,
+}
+
+impl<'a> Drop for BoolResetter<'a> {
+    fn drop(&mut self) {
+        self.inner.store(self.reset_to, self.ordering);
+    }
+}
+
+impl<'a> BoolResetter<'a> {
+    /// Disarm the guard, consuming it to make it not reset any more.
+    fn disarm(self) {
+        std::mem::forget(self);
+    }
+}
+
+/// The possible origins of a document.
+///
+/// Used (for example) to report where we got a document from if it fails to
+/// parse.
+#[derive(Debug, Clone, derive_more::Display)]
+#[non_exhaustive]
+pub enum DocSource {
+    /// We loaded the document from our cache.
+    #[display(fmt = "local cache")]
+    LocalCache,
+    /// We fetched the document from a server.
+    //
+    // TODO: we'll should add a lot more information here in the future, once
+    // it's available from tor_dirclient::DirSource,
+    #[display(fmt = "directory server")]
+    DirServer {},
 }
 
 impl<R: Runtime> DirMgr<R> {
@@ -168,30 +299,75 @@ impl<R: Runtime> DirMgr<R> {
         circmgr: Arc<CircMgr<R>>,
     ) -> Result<Arc<NetDir>> {
         let dirmgr = DirMgr::bootstrap_from_config(config, runtime, circmgr).await?;
-        Ok(dirmgr.netdir())
+        dirmgr.netdir()
     }
 
-    /// Return a new directory manager from a given configuration,
-    /// bootstrapping from the network as necessary.
+    /// Create a new `DirMgr` in online mode, but don't bootstrap it yet.
     ///
-    /// This function will to return until the directory is
-    /// bootstrapped enough to build circuits.  It will also launch a
-    /// background task that fetches any missing information, and that
-    /// replaces the directory when a new one is available.
-    pub async fn bootstrap_from_config(
+    /// The `DirMgr` can be bootstrapped later with `bootstrap`.
+    pub fn create_unbootstrapped(
         config: DirMgrConfig,
         runtime: R,
         circmgr: Arc<CircMgr<R>>,
     ) -> Result<Arc<Self>> {
-        let dirmgr = Arc::new(DirMgr::from_config(
+        Ok(Arc::new(DirMgr::from_config(
             config,
-            runtime.clone(),
+            runtime,
             Some(circmgr),
             false,
-        )?);
+        )?))
+    }
+
+    /// Bootstrap a `DirMgr` created in online mode that hasn't been bootstrapped yet.
+    ///
+    /// This function will not return until the directory is bootstrapped enough to build circuits.
+    /// It will also launch a background task that fetches any missing information, and that
+    /// replaces the directory when a new one is available.
+    ///
+    /// This function is intended to be used together with `create_unbootstrapped`. There is no
+    /// need to call this function otherwise.
+    ///
+    /// If bootstrapping has already successfully taken place, returns early with success.
+    ///
+    /// # Errors
+    ///
+    /// Returns an error if bootstrapping fails. If the error is [`Error::CantAdvanceState`],
+    /// it may be possible to successfully bootstrap later on by calling this function again.
+    ///
+    /// # Panics
+    ///
+    /// Panics if the `DirMgr` passed to this function was not created in online mode, such as
+    /// via `load_once`.
+    pub async fn bootstrap(self: &Arc<Self>) -> Result<()> {
+        if self.offline {
+            return Err(Error::OfflineMode);
+        }
+
+        // The semantics of this are "attempt to replace a 'false' value with 'true'.
+        // If the value in bootstrap_started was not 'false' when the attempt was made, returns
+        // `Err`; this means another bootstrap attempt is in progress or has completed, so we
+        // return early.
+
+        // NOTE(eta): could potentially weaken the `Ordering` here in future
+        if self
+            .bootstrap_started
+            .compare_exchange(false, true, Ordering::SeqCst, Ordering::SeqCst)
+            .is_err()
+        {
+            debug!("Attempted to bootstrap twice; ignoring.");
+            return Ok(());
+        }
+
+        // Use a RAII guard to reset `bootstrap_started` to `false` if we return early without
+        // completing bootstrap.
+        let resetter = BoolResetter {
+            inner: &self.bootstrap_started,
+            reset_to: false,
+            ordering: Ordering::SeqCst,
+        };
 
         // Try to load from the cache.
-        let have_directory = dirmgr.load_directory().await?;
+        let have_directory = self.load_directory().await?;
 
         let (mut sender, receiver) = if have_directory {
             info!("Loaded a good directory from cache.");
@@ -203,30 +379,34 @@ impl<R: Runtime> DirMgr<R> {
         };
 
         // Whether we loaded or not, we now start downloading.
-        let dirmgr_weak = Arc::downgrade(&dirmgr);
-        runtime.spawn(async move {
-            // NOTE: This is a daemon task.  It should eventually get
-            // treated as one.
+        let dirmgr_weak = Arc::downgrade(self);
+        self.runtime
+            .spawn(async move {
+                // NOTE: This is a daemon task.  It should eventually get
+                // treated as one.
 
-            // Don't warn when these are Error::ManagerDropped: that
-            // means that the DirMgr has been shut down.
-            if let Err(e) = Self::reload_until_owner(&dirmgr_weak, &mut sender).await {
-                match e {
-                    Error::ManagerDropped => {}
-                    _ => warn!("Unrecovered error while waiting for bootstrap: {}", e),
+                // Don't warn when these are Error::ManagerDropped: that
+                // means that the DirMgr has been shut down.
+                if let Err(e) = Self::reload_until_owner(&dirmgr_weak, &mut sender).await {
+                    match e {
+                        Error::ManagerDropped => {}
+                        _ => warn!("Unrecovered error while waiting for bootstrap: {}", e),
+                    }
+                } else if let Err(e) = Self::download_forever(dirmgr_weak, sender).await {
+                    match e {
+                        Error::ManagerDropped => {}
+                        _ => warn!("Unrecovered error while downloading: {}", e),
+                    }
                 }
-            } else if let Err(e) = Self::download_forever(dirmgr_weak, sender).await {
-                match e {
-                    Error::ManagerDropped => {}
-                    _ => warn!("Unrecovered error while downloading: {}", e),
-                }
-            }
-        })?;
+            })
+            .map_err(|e| Error::from_spawn("directory updater task", e))?;
 
         if let Some(receiver) = receiver {
             match receiver.await {
                 Ok(()) => {
                     info!("We have enough information to build circuits.");
+                    // Disarm the RAII guard, since we succeeded.
+                    resetter.disarm();
                 }
                 Err(_) => {
                     warn!("Bootstrapping task exited before finishing.");
@@ -234,6 +414,24 @@ impl<R: Runtime> DirMgr<R> {
                 }
             }
         }
+        Ok(())
+    }
+
+    /// Returns `true` if a bootstrap attempt is in progress, or successfully completed.
+    pub fn bootstrap_started(&self) -> bool {
+        self.bootstrap_started.load(Ordering::SeqCst)
+    }
+
+    /// Return a new directory manager from a given configuration,
+    /// bootstrapping from the network as necessary.
+    pub async fn bootstrap_from_config(
+        config: DirMgrConfig,
+        runtime: R,
+        circmgr: Arc<CircMgr<R>>,
+    ) -> Result<Arc<Self>> {
+        let dirmgr = Self::create_unbootstrapped(config, runtime, circmgr)?;
+
+        dirmgr.bootstrap().await?;
 
         Ok(dirmgr)
     }
@@ -423,7 +621,7 @@ impl<R: Runtime> DirMgr<R> {
         let params_changed = new_config.override_net_params() != config.override_net_params();
 
         self.config
-            .map_and_replace(|cfg| cfg.update_config(new_config));
+            .map_and_replace(|cfg| cfg.update_from_config(new_config));
 
         if params_changed {
             let _ignore_err = self.netdir.mutate(|netdir| {
@@ -438,6 +636,25 @@ impl<R: Runtime> DirMgr<R> {
         Ok(())
     }
 
+    /// Return a stream of [`DirBootstrapStatus`] events to tell us about changes
+    /// in the latest directory's bootstrap status.
+    ///
+    /// Note that this stream can be lossy: the caller will not necessarily
+    /// observe every event on the stream
+    pub fn bootstrap_events(&self) -> event::DirBootstrapEvents {
+        self.receive_status.clone()
+    }
+
+    /// Replace the latest status with `new_status` and broadcast to anybody
+    /// watching via a [`DirBootstrapEvents`] stream.
+    fn update_status(&self, new_status: DirStatus) {
+        // TODO(nickm): can I kill off this lock by having something else own the sender?
+        let mut sender = self.send_status.lock().expect("poisoned lock");
+        let mut status = sender.borrow_mut();
+
+        status.update(new_status);
+    }
+
     /// Try to make this a directory manager with read-write access to its
     /// storage.
     ///
@@ -452,7 +669,7 @@ impl<R: Runtime> DirMgr<R> {
     }
 
     /// Return a reference to the store, if it is currently read-write.
-    fn store_if_rw(&self) -> Option<&Mutex<SqliteStore>> {
+    fn store_if_rw(&self) -> Option<&Mutex<DynStore>> {
         let rw = !self
             .store
             .lock()
@@ -467,23 +684,40 @@ impl<R: Runtime> DirMgr<R> {
     }
 
     /// Construct a DirMgr from a DirMgrConfig.
+    ///
+    /// If `offline` is set, opens the SQLite store read-only and sets the offline flag in the
+    /// returned manager.
     fn from_config(
         config: DirMgrConfig,
         runtime: R,
         circmgr: Option<Arc<CircMgr<R>>>,
-        readonly: bool,
+        offline: bool,
     ) -> Result<Self> {
-        let store = Mutex::new(config.open_sqlite_store(readonly)?);
+        let store = Mutex::new(config.open_store(offline)?);
         let netdir = SharedMutArc::new();
         let events = event::FlagPublisher::new();
 
+        let (send_status, receive_status) = postage::watch::channel();
+        let send_status = Mutex::new(send_status);
+        let receive_status = DirBootstrapEvents {
+            inner: receive_status,
+        };
+        #[cfg(feature = "dirfilter")]
+        let filter = config.extensions.filter.clone();
+
         Ok(DirMgr {
             config: config.into(),
             store,
             netdir,
             events,
+            send_status,
+            receive_status,
             circmgr,
             runtime,
+            offline,
+            bootstrap_started: AtomicBool::new(false),
+            #[cfg(feature = "dirfilter")]
+            filter,
         })
     }
 
@@ -499,18 +733,18 @@ impl<R: Runtime> DirMgr<R> {
     }
 
     /// Return an Arc handle to our latest directory, if we have one.
-    ///
-    /// This is a private method, since by the time anybody else has a
-    /// handle to a DirMgr, the NetDir should definitely be
-    /// bootstrapped.
-    fn opt_netdir(&self) -> Option<Arc<NetDir>> {
+    pub fn opt_netdir(&self) -> Option<Arc<NetDir>> {
         self.netdir.get()
     }
 
-    /// Return an Arc handle to our latest directory, if we have one.
+    /// Return an Arc handle to our latest directory, returning an error if there is none.
+    ///
+    /// # Errors
+    ///
+    /// Errors with [`Error::DirectoryNotPresent`] if the `DirMgr` hasn't been bootstrapped yet.
     // TODO: Add variants of this that make sure that it's up-to-date?
-    pub fn netdir(&self) -> Arc<NetDir> {
-        self.opt_netdir().expect("DirMgr was not bootstrapped!")
+    pub fn netdir(&self) -> Result<Arc<NetDir>> {
+        self.opt_netdir().ok_or(Error::DirectoryNotPresent)
     }
 
     /// Return a new asynchronous stream that will receive notification
@@ -623,7 +857,7 @@ impl<R: Runtime> DirMgr<R> {
         for q in q.split_for_download() {
             match q {
                 DocQuery::LatestConsensus { flavor, .. } => {
-                    res.push(self.make_consensus_request(flavor)?);
+                    res.push(self.make_consensus_request(self.runtime.wallclock(), flavor)?);
                 }
                 DocQuery::AuthCert(ids) => {
                     res.push(ClientRequest::AuthCert(ids.into_iter().collect()));
@@ -642,19 +876,30 @@ impl<R: Runtime> DirMgr<R> {
 
     /// Construct an appropriate ClientRequest to download a consensus
     /// of the given flavor.
-    fn make_consensus_request(&self, flavor: ConsensusFlavor) -> Result<ClientRequest> {
-        #![allow(clippy::unnecessary_wraps)]
+    fn make_consensus_request(
+        &self,
+        now: SystemTime,
+        flavor: ConsensusFlavor,
+    ) -> Result<ClientRequest> {
         let mut request = tor_dirclient::request::ConsensusRequest::new(flavor);
 
+        let default_cutoff = default_consensus_cutoff(now)?;
+
         let r = self.store.lock().expect("Directory storage lock poisoned");
         match r.latest_consensus_meta(flavor) {
             Ok(Some(meta)) => {
-                request.set_last_consensus_date(meta.lifetime().valid_after());
+                let valid_after = meta.lifetime().valid_after();
+                request.set_last_consensus_date(std::cmp::max(valid_after, default_cutoff));
                 request.push_old_consensus_digest(*meta.sha3_256_of_signed());
             }
-            Ok(None) => {}
-            Err(e) => {
-                warn!("Error loading directory metadata: {}", e);
+            latest => {
+                if let Err(e) = latest {
+                    warn!("Error loading directory metadata: {}", e);
+                }
+                // If we don't have a consensus, then request one that's
+                // "reasonably new".  That way, our clock is set far in the
+                // future, we won't download stuff we can't use.
+                request.set_last_consensus_date(default_cutoff);
             }
         }
 
@@ -694,6 +939,70 @@ impl<R: Runtime> DirMgr<R> {
         }
         Ok(text)
     }
+
+    /// If there were errors from a peer in `outcome`, record those errors by
+    /// marking the circuit (if any) as needing retirement, and noting the peer
+    /// (if any) as having failed.
+    fn note_request_outcome(&self, outcome: &tor_dirclient::Result<tor_dirclient::DirResponse>) {
+        use tor_dirclient::Error::RequestFailed;
+        // Extract an error and a source from this outcome, if there is one.
+        //
+        // This is complicated because DirResponse can encapsulate the notion of
+        // a response that failed part way through a download: in the case, it
+        // has some data, and also an error.
+        let (err, source) = match outcome {
+            Ok(req) => {
+                if let (Some(e), Some(source)) = (req.error(), req.source()) {
+                    (
+                        RequestFailed {
+                            error: e.clone(),
+                            source: Some(source.clone()),
+                        },
+                        source,
+                    )
+                } else {
+                    return;
+                }
+            }
+            Err(RequestFailed {
+                source: Some(source),
+                ..
+            }) => (
+                // TODO: Use an @ binding in the pattern once we are on MSRV >=
+                // 1.56.
+                outcome.as_ref().unwrap_err().clone(),
+                source,
+            ),
+            _ => return,
+        };
+
+        self.note_cache_error(source, &err.into());
+    }
+
+    /// Record that a problem has occurred because of a failure in an answer from `source`.
+    fn note_cache_error(&self, source: &tor_dirclient::SourceInfo, problem: &Error) {
+        use tor_circmgr::ExternalActivity;
+
+        if !problem.indicates_cache_failure() {
+            return;
+        }
+
+        if let Some(circmgr) = &self.circmgr {
+            info!("Marking {:?} as failed: {}", source, problem);
+            circmgr.note_external_failure(source.cache_id(), ExternalActivity::DirCache);
+            circmgr.retire_circ(source.unique_circ_id());
+        }
+    }
+
+    /// Record that `source` has successfully given us some directory info.
+    fn note_cache_success(&self, source: &tor_dirclient::SourceInfo) {
+        use tor_circmgr::ExternalActivity;
+
+        if let Some(circmgr) = &self.circmgr {
+            trace!("Marking {:?} as successful", source);
+            circmgr.note_external_success(source.cache_id(), ExternalActivity::DirCache);
+        }
+    }
 }
 
 /// A degree of readiness for a given directory state object.
@@ -744,7 +1053,7 @@ trait DirState: Send {
     fn add_from_cache(
         &mut self,
         docs: HashMap<DocId, DocumentText>,
-        storage: Option<&Mutex<SqliteStore>>,
+        storage: Option<&Mutex<DynStore>>,
     ) -> Result<bool>;
 
     /// Add information that we have just downloaded to this state; returns
@@ -764,8 +1073,11 @@ trait DirState: Send {
         &mut self,
         text: &str,
         request: &ClientRequest,
-        storage: Option<&Mutex<SqliteStore>>,
+        storage: Option<&Mutex<DynStore>>,
     ) -> Result<bool>;
+    /// Return a summary of this state as a [`DirStatus`].
+    fn bootstrap_status(&self) -> event::DirStatus;
+
     /// Return a configuration for attempting downloads.
     fn dl_config(&self) -> Result<DownloadSchedule>;
     /// If possible, advance to the next state.
@@ -783,6 +1095,31 @@ fn upgrade_weak_ref<T>(weak: &Weak<T>) -> Result<Arc<T>> {
     Weak::upgrade(weak).ok_or(Error::ManagerDropped)
 }
 
+/// At most how much age can we tolerate in a consensus?
+///
+/// TODO: Make this public and/or use it elsewhere; see arti#412.
+const CONSENSUS_ALLOW_SKEW: Duration = Duration::from_secs(3600 * 48);
+
+/// Given a time `now`, return the age of the oldest consensus that we should
+/// request at that time.
+fn default_consensus_cutoff(now: SystemTime) -> Result<SystemTime> {
+    let cutoff = time::OffsetDateTime::from(now - CONSENSUS_ALLOW_SKEW);
+    // We now round cutoff to the next hour, so that we aren't leaking our exact
+    // time to the directory cache.
+    //
+    // With the time crate, it's easier to calculate the "next hour" by rounding
+    // _down_ then adding an hour; rounding up would sometimes require changing
+    // the date too.
+    let (h, _m, _s) = cutoff.to_hms();
+    let cutoff = cutoff.replace_time(
+        time::Time::from_hms(h, 0, 0)
+            .map_err(tor_error::into_internal!("Failed clock calculation"))?,
+    );
+    let cutoff = cutoff + Duration::from_secs(3600);
+
+    Ok(cutoff.into())
+}
+
 #[cfg(test)]
 mod test {
     #![allow(clippy::unwrap_used)]
@@ -791,13 +1128,14 @@ mod test {
     use std::time::Duration;
     use tempfile::TempDir;
     use tor_netdoc::doc::{authcert::AuthCertKeyIds, netstatus::Lifetime};
+    use tor_rtcompat::SleepProvider;
 
     pub(crate) fn new_mgr<R: Runtime>(runtime: R) -> (TempDir, DirMgr<R>) {
         let dir = TempDir::new().unwrap();
-        let config = DirMgrConfig::builder()
-            .cache_path(dir.path())
-            .build()
-            .unwrap();
+        let config = DirMgrConfig {
+            cache_path: dir.path().into(),
+            ..Default::default()
+        };
         let dirmgr = DirMgr::from_config(config, runtime, None, false).unwrap();
 
         (dir, dirmgr)
@@ -816,12 +1154,12 @@ mod test {
     #[test]
     fn load_and_store_internals() {
         tor_rtcompat::test_with_one_runtime!(|rt| async {
-            let (_tempdir, mgr) = new_mgr(rt);
-
-            let now = SystemTime::now();
+            let now = rt.wallclock();
             let tomorrow = now + Duration::from_secs(86400);
             let later = tomorrow + Duration::from_secs(86400);
 
+            let (_tempdir, mgr) = new_mgr(rt);
+
             // Seed the storage with a bunch of junk.
             let d1 = [5_u8; 32];
             let d2 = [7; 32];
@@ -842,7 +1180,7 @@ mod test {
 
                 store
                     .store_microdescs(
-                        vec![
+                        &[
                             ("Fake micro 1", &d1),
                             ("Fake micro 2", &d2),
                             ("Fake micro 3", &d3),
@@ -853,7 +1191,7 @@ mod test {
 
                 #[cfg(feature = "routerdesc")]
                 store
-                    .store_routerdescs(vec![("Fake rd1", now, &d4), ("Fake rd2", now, &d5)])
+                    .store_routerdescs(&[("Fake rd1", now, &d4), ("Fake rd2", now, &d5)])
                     .unwrap();
 
                 store
@@ -931,20 +1269,22 @@ mod test {
     #[test]
     fn make_consensus_request() {
         tor_rtcompat::test_with_one_runtime!(|rt| async {
-            let (_tempdir, mgr) = new_mgr(rt);
-
-            let now = SystemTime::now();
+            let now = rt.wallclock();
             let tomorrow = now + Duration::from_secs(86400);
             let later = tomorrow + Duration::from_secs(86400);
 
+            let (_tempdir, mgr) = new_mgr(rt);
+
             // Try with an empty store.
             let req = mgr
-                .make_consensus_request(ConsensusFlavor::Microdesc)
+                .make_consensus_request(now, ConsensusFlavor::Microdesc)
                 .unwrap();
             match req {
                 ClientRequest::Consensus(r) => {
                     assert_eq!(r.old_consensus_digests().count(), 0);
-                    assert_eq!(r.last_consensus_date(), None);
+                    let date = r.last_consensus_date().unwrap();
+                    assert!(date >= now - CONSENSUS_ALLOW_SKEW);
+                    assert!(date <= now - CONSENSUS_ALLOW_SKEW + Duration::from_secs(3600));
                 }
                 _ => panic!("Wrong request type"),
             }
@@ -966,7 +1306,7 @@ mod test {
 
             // Now try again.
             let req = mgr
-                .make_consensus_request(ConsensusFlavor::Microdesc)
+                .make_consensus_request(now, ConsensusFlavor::Microdesc)
                 .unwrap();
             match req {
                 ClientRequest::Consensus(r) => {
@@ -1026,6 +1366,9 @@ mod test {
     #[test]
     fn expand_response() {
         tor_rtcompat::test_with_one_runtime!(|rt| async {
+            let now = rt.wallclock();
+            let day = Duration::from_secs(86400);
+
             let (_tempdir, mgr) = new_mgr(rt);
 
             // Try a simple request: nothing should happen.
@@ -1049,8 +1392,6 @@ mod test {
             // we can ask for a diff.
             {
                 let mut store = mgr.store.lock().unwrap();
-                let now = SystemTime::now();
-                let day = Duration::from_secs(86400);
                 let d_in = [0x99; 32]; // This one, we can fake.
                 let cmeta = ConsensusMeta::new(
                     Lifetime::new(now, now + day, now + 2 * day).unwrap(),
@@ -1095,4 +1436,31 @@ replacement line
             assert!(expanded.is_err());
         });
     }
+
+    #[test]
+    #[allow(clippy::bool_assert_comparison)]
+    fn bool_resetter_works() {
+        let bool = AtomicBool::new(false);
+        fn example(bool: &AtomicBool) {
+            bool.store(true, Ordering::SeqCst);
+            let _resetter = BoolResetter {
+                inner: bool,
+                reset_to: false,
+                ordering: Ordering::SeqCst,
+            };
+        }
+        fn example_disarm(bool: &AtomicBool) {
+            bool.store(true, Ordering::SeqCst);
+            let resetter = BoolResetter {
+                inner: bool,
+                reset_to: false,
+                ordering: Ordering::SeqCst,
+            };
+            resetter.disarm();
+        }
+        example(&bool);
+        assert_eq!(bool.load(Ordering::SeqCst), false);
+        example_disarm(&bool);
+        assert_eq!(bool.load(Ordering::SeqCst), true);
+    }
 }
