@@ -84,6 +84,10 @@ pub(crate) trait WriteNetDir: 'static + Sync + Send {
     /// testing it is helpful to be able to mock our our current view
     /// of the time.
     fn now(&self) -> SystemTime;
+
+    /// Return the currently configured DynFilter for this state.
+    #[cfg(feature = "dirfilter")]
+    fn filter(&self) -> &dyn crate::filter::DirFilter;
 }
 
 impl<R: Runtime> WriteNetDir for crate::DirMgr<R> {
@@ -106,7 +110,12 @@ impl<R: Runtime> WriteNetDir for crate::DirMgr<R> {
         }
     }
     fn now(&self) -> SystemTime {
-        SystemTime::now()
+        self.runtime.wallclock()
+    }
+
+    #[cfg(feature = "dirfilter")]
+    fn filter(&self) -> &dyn crate::filter::DirFilter {
+        self.filter.as_deref().unwrap_or(&crate::filter::NilFilter)
     }
 }
 
@@ -149,7 +158,7 @@ impl<DM: WriteNetDir> GetConsensusState<DM> {
                 .config()
                 .authorities()
                 .iter()
-                .map(|auth| *auth.v3ident())
+                .map(|auth| auth.v3ident)
                 .collect();
             let after = writedir
                 .netdir()
@@ -278,6 +287,12 @@ impl<DM: WriteNetDir> GetConsensusState<DM> {
         let (consensus_meta, unvalidated) = {
             let (signedval, remainder, parsed) =
                 MdConsensus::parse(text).map_err(|e| Error::from_netdoc(source.clone(), e))?;
+            #[cfg(feature = "dirfilter")]
+            let parsed = if let Some(wd) = Weak::upgrade(&self.writedir) {
+                wd.filter().filter_consensus(parsed)?
+            } else {
+                parsed
+            };
             let now = current_time(&self.writedir)?;
             if let Ok(timely) = parsed.check_valid_at(&now) {
                 let meta = ConsensusMeta::from_unvalidated(signedval, remainder, &timely);
@@ -518,8 +533,6 @@ impl<DM: WriteNetDir> DirState for GetCertsState<DM> {
 struct GetMicrodescsState<DM: WriteNetDir> {
     /// How should we get the consensus from the cache, if at all?
     cache_usage: CacheUsage,
-    /// The digests of the microdescriptors we are missing.
-    missing: HashSet<MdDigest>,
     /// Total number of microdescriptors listed in the consensus.
     n_microdescs: usize,
     /// The dirmgr to inform about a usable directory.
@@ -555,10 +568,14 @@ enum PendingNetDir {
     WaitingForGuards(NetDir),
 }
 
-impl PendingNetDir {
-    /// Add the provided microdescriptor to this pending directory.
-    ///
-    /// Return true if we indeed wanted (and added) this descriptor.
+impl MdReceiver for PendingNetDir {
+    fn missing_microdescs(&self) -> Box<dyn Iterator<Item = &MdDigest> + '_> {
+        match self {
+            PendingNetDir::Partial(partial) => partial.missing_microdescs(),
+            PendingNetDir::WaitingForGuards(netdir) => netdir.missing_microdescs(),
+        }
+    }
+
     fn add_microdesc(&mut self, md: Microdesc) -> bool {
         match self {
             PendingNetDir::Partial(partial) => partial.add_microdesc(md),
@@ -566,6 +583,15 @@ impl PendingNetDir {
         }
     }
 
+    fn n_missing(&self) -> usize {
+        match self {
+            PendingNetDir::Partial(partial) => partial.n_missing(),
+            PendingNetDir::WaitingForGuards(netdir) => netdir.n_missing(),
+        }
+    }
+}
+
+impl PendingNetDir {
     /// Try to move `self` as far as possible towards a complete, netdir with
     /// enough directory information (according to `writedir`).
     ///
@@ -618,11 +644,9 @@ impl<DM: WriteNetDir> GetMicrodescsState<DM> {
             None => return Err(Error::ManagerDropped),
         };
 
-        let missing = partial_dir.missing_microdescs().map(Clone::clone).collect();
         let mut result = GetMicrodescsState {
             cache_usage,
             n_microdescs,
-            missing,
             writedir,
             partial: Some(PendingNetDir::Partial(partial_dir)),
             meta,
@@ -642,6 +666,14 @@ impl<DM: WriteNetDir> GetMicrodescsState<DM> {
     where
         I: IntoIterator<Item = Microdesc>,
     {
+        #[cfg(feature = "dirfilter")]
+        let mds: Vec<Microdesc> = if let Some(wd) = Weak::upgrade(&self.writedir) {
+            mds.into_iter()
+                .filter_map(|m| wd.filter().filter_md(m).ok())
+                .collect()
+        } else {
+            mds.into_iter().collect()
+        };
         if let Some(p) = &mut self.partial {
             for md in mds {
                 self.newly_listed.push(*md.digest());
@@ -703,19 +735,52 @@ impl<DM: WriteNetDir> GetMicrodescsState<DM> {
     }
 }
 
+impl<DM: WriteNetDir> GetMicrodescsState<DM> {
+    /// Try to obtain info from an inner `MdReceiver`
+    ///
+    /// Either finds an inner `MdReceiver` and calls `f` on it, or returns `default()`.
+    ///
+    /// Used for missing microdescs.
+    fn with_mdreceiver_for_missing<F, T>(&self, f: F) -> T
+    where
+        F: FnOnce(&dyn MdReceiver) -> T,
+        T: Default,
+    {
+        if let Some(partial) = &self.partial {
+            return f(partial);
+        } else if let Some(wd) = Weak::upgrade(&self.writedir) {
+            if let Some(nd) = wd.netdir().get() {
+                return f(nd.as_ref());
+            }
+        }
+        Default::default()
+    }
+
+    /// Number of missing microdescriptors, or 0
+    ///
+    /// Can return 0 if we don't have that information.
+    fn n_missing(&self) -> usize {
+        self.with_mdreceiver_for_missing(|d| d.n_missing())
+    }
+}
+
 impl<DM: WriteNetDir> DirState for GetMicrodescsState<DM> {
     fn describe(&self) -> String {
         format!(
             "Downloading microdescriptors (we are missing {}).",
-            self.missing.len()
+            self.n_missing()
         )
     }
     fn missing_docs(&self) -> Vec<DocId> {
-        self.missing.iter().map(|d| DocId::Microdesc(*d)).collect()
+        self.with_mdreceiver_for_missing(|d| {
+            d.missing_microdescs()
+                .map(|d| DocId::Microdesc(*d))
+                .collect()
+        })
     }
     fn is_ready(&self, ready: Readiness) -> bool {
         match ready {
-            Readiness::Complete => self.missing.is_empty(),
+            Readiness::Complete => self.n_missing() == 0,
             Readiness::Usable => self.partial.is_none(),
         }
     }
@@ -723,7 +788,7 @@ impl<DM: WriteNetDir> DirState for GetMicrodescsState<DM> {
         false
     }
     fn bootstrap_status(&self) -> DirStatus {
-        let n_present = self.n_microdescs - self.missing.len();
+        let n_present = self.n_microdescs - self.n_missing();
         DirStatusInner::Validated {
             lifetime: self.meta.lifetime().clone(),
             n_mds: (n_present as u32, self.n_microdescs as u32),
@@ -746,10 +811,6 @@ impl<DM: WriteNetDir> DirState for GetMicrodescsState<DM> {
         let mut microdescs = Vec::new();
         for (id, text) in docs {
             if let DocId::Microdesc(digest) = id {
-                if !self.missing.remove(&digest) {
-                    warn!("Bug: loaded a microdesc that we didn't want from the cache.");
-                    continue;
-                }
                 if let Ok(md) = Microdesc::parse(text.as_str().map_err(Error::BadUtf8InCache)?) {
                     if md.digest() == &digest {
                         microdescs.push(md);
@@ -764,6 +825,10 @@ impl<DM: WriteNetDir> DirState for GetMicrodescsState<DM> {
         if self.register_microdescs(microdescs) {
             // Just stopped being pending.
             self.mark_consensus_usable(storage)?;
+            // We can save a lot of ram this way, though we don't want to do it
+            // so often.  As a compromise we call `shrink_to_fit at most twice
+            // per consensus: once when we're ready to use, and once when we are
+            // complete.
         }
 
         Ok(changed)
@@ -793,7 +858,6 @@ impl<DM: WriteNetDir> DirState for GetMicrodescsState<DM> {
                 );
                 continue;
             }
-            self.missing.remove(md.digest());
             new_mds.push((txt, md));
         }
 
@@ -959,11 +1023,11 @@ mod test {
             if let Some(a) = authorities {
                 netcfg.authorities(a);
             }
-            let cfg = DirMgrConfig::builder()
-                .cache_path("/we_will_never_use_this/")
-                .network_config(netcfg.build().unwrap())
-                .build()
-                .unwrap();
+            let cfg = DirMgrConfig {
+                cache_path: "/we_will_never_use_this/".into(),
+                network_config: netcfg.build().unwrap(),
+                ..Default::default()
+            };
             let cfg = Arc::new(cfg);
             DirRcv {
                 now,
@@ -992,6 +1056,10 @@ mod test {
         fn now(&self) -> SystemTime {
             self.now
         }
+        #[cfg(feature = "dirfilter")]
+        fn filter(&self) -> &dyn crate::filter::DirFilter {
+            &crate::filter::NilFilter
+        }
     }
 
     // Test data
@@ -1005,8 +1073,7 @@ mod test {
         datetime!(2020-08-07 12:42:45 UTC).into()
     }
     fn rsa(s: &str) -> RsaIdentity {
-        let k = hex::decode(s).unwrap();
-        RsaIdentity::from_bytes(&k[..]).unwrap()
+        RsaIdentity::from_hex(s).unwrap()
     }
     fn test_authorities() -> Vec<Authority> {
         fn a(s: &str) -> Authority {
